"""Data object for verse ranges."""

from typing import List, Self, Tuple
from pydantic import BaseModel, model_validator

from refspy.number import Number
from refspy.verse import verse, Verse


class Range(BaseModel):
    start: Verse
    end: Verse

    def tuple(self) -> Tuple[Verse, Verse]:
        return (self.start, self.end)

    @model_validator(mode="after")
    def check_verse_order(self) -> Self:
        """Enforce verse order within the Range.

        Raises:
            ValueError: If the start verse is greater than the end verse.
        """
        assert tuple(self.start) <= tuple(self.end)
        return self

    def __lt__(self, other: Self) -> bool:
        return tuple(self) < tuple(other)

    def overlaps(self, other: Self) -> bool:
        """Determine whether this reference overlaps the other.

        Overlap is when either's start or end falls within the other's range.

        Used in `refspy.reference.reference.overlaps()`
        """
        return any(
            [
                other.start <= self.start <= other.end,
                other.start <= self.end <= other.end,
                self.start <= other.start <= self.end,
                self.start <= other.end <= self.end,
            ]
        )

    def contains(self, other: Self) -> bool:
        """Determine whether this reference contains the other.

        Containment is when this reference's range includes the other's start
        and end.

        Used in `refspy.reference.reference.contains()`
        """
        return self.start <= other.start and self.end >= other.end

    def adjoins(self, other: Self) -> bool:
        """Determine whether this reference is adjacent to the other.

        Adjacency determines whether two ranges can be joined by
        `refspy.ranges.join_ranges()`.

        To keep things simple while covering most normal cases, adjacency is
        when:

        * Two verse ranges are adjacent.
        * Two inter-chapter verse ranges are adjacent.
        * Two chapter_range references are adjacent.
        * Two book_range references are adjacent.

        Note:
            We do not calculate adjacency for inter-book or inter-chapter
            references.
        """
        if any(
            [self.is_verse(), self.is_verse_range(), self.is_inter_chapter_range()]
        ) and any(
            [other.is_verse(), other.is_verse_range(), other.is_inter_chapter_range()]
        ):
            return (
                other.start.verse == self.end.verse + 1
                or self.start.verse == other.end.verse + 1
            )
        elif (
            self.same_book_as(other)
            and (self.is_chapter() or self.is_chapter_range())
            and (other.is_chapter() or other.is_chapter_range())
        ):
            return (
                other.start.chapter == self.end.chapter + 1
                or self.start.chapter == other.end.chapter + 1
            )
        elif (
            self.same_library_as(other)
            and (self.is_book() or self.is_book_range())
            and (other.is_book() or other.is_book_range())
        ):
            return (
                other.start.book == self.end.book + 1
                or self.start.book == other.end.book + 1
            )
        else:
            return False

    def merge(self, other: Self) -> Self:
        """Combine two overlapping ranges."""
        min_start_c, min_start_v = min(
            (self.start.chapter, self.start.verse),
            (other.start.chapter, other.start.verse),
        )
        max_end_c, max_end_v = max(
            (self.end.chapter, self.end.verse),
            (other.end.chapter, other.end.verse),
        )
        return self.__class__(
            start=verse(
                min(self.start.library, other.start.library),
                min(self.start.book, other.start.book),
                min_start_c,
                min_start_v,
            ),
            end=verse(
                max(self.end.library, other.end.library),
                max(self.end.book, other.end.book),
                max_end_c,
                max_end_v,
            ),
        )

    def join(self, other: Self):
        """Combine two adjacent ranges."""
        if self.start < other.start:
            return self.__class__(start=self.start, end=other.end)
        else:
            return self.__class__(start=other.start, end=self.end)

    def is_same_library(self) -> bool:
        return self.start.library == self.end.library

    def same_library_as(self, other: Self) -> bool:
        """Determine if two ranges are in the same library."""
        ids = set(
            [
                self.start.library,
                self.end.library,
                other.start.library,
                other.end.library,
            ]
        )
        return len(ids) == 1

    def is_same_book(self) -> bool:
        return self.is_same_library() and self.start.book == self.end.book

    def same_book_as(self, other: Self) -> bool:
        """Determine if two references are in the same book.

        To be in the same book they must also be in the same library.
        """
        ids = set([self.start.book, self.end.book, other.start.book, other.end.book])
        return self.same_library_as(other) and len(ids) == 1

    def is_same_chapter(self) -> bool:
        return (
            self.is_same_library()
            and self.is_same_book()
            and self.start.chapter == self.end.chapter
        )

    def same_chapter_as(self, other: Self) -> bool:
        """Determine if two references are in the same book.

        To be in the same book they must also be in the same library.
        """
        ids = {
            self.start.chapter,
            self.end.chapter,
            other.start.chapter,
            other.end.chapter,
        }
        return self.same_book_as(other) and len(ids) == 1

    def is_book(self) -> bool:
        """Determine if this range covers one whole book.

        Example:
            `Matthew`
        """
        return all(
            [
                self.start.library == self.end.library,
                self.start.book == self.end.book,
                (self.start.chapter, self.end.chapter) == (1, 999),
                (self.start.verse, self.end.verse) == (1, 999),
            ]
        )

    def is_book_range(self) -> bool:
        """Determine if this range goes from one whole book to another whole book.

        Example:
            `Matthew-John`
        """
        return all(
            [
                self.start.library == self.end.library,
                self.start.book != self.end.book,
                (self.start.chapter, self.end.chapter) == (1, 999),
                (self.start.verse, self.end.verse) == (1, 999),
            ]
        )

    def is_inter_book_range(self) -> bool:
        """Determine if this range spans multiple books.

        Example:
            `Matt 1:2-Mark 3:4`
        """
        return all(
            [
                self.start.library == self.end.library,
                self.start.book != self.end.book,
                any(
                    [
                        self.start.chapter != 1,
                        self.end.chapter != 999,
                    ]
                ),
            ]
        )

    def is_chapter(self) -> bool:
        """Determine if this range covers one whole chapter.

        Example:
            `Matthew 7`
        """
        return all(
            [
                self.start.library == self.end.library,
                self.start.book == self.end.book,
                self.start.chapter == self.end.chapter,
                (self.start.verse, self.end.verse) == (1, 999),
            ]
        )

    def is_chapter_range(self) -> bool:
        """Determine if this range goes from one whole chapter to another whole
        chapter.

        Example:
            `Matthew 5-8`
        """
        return all(
            [
                self.start.library == self.end.library,
                self.start.book == self.end.book,
                self.start.chapter != self.end.chapter,
                (self.start.verse, self.end.verse) == (1, 999),
            ]
        )

    def is_inter_chapter_range(self) -> bool:
        """Determine if this range spans multiple chapters within the same book.

        Example:
            `Matthew 1:2-3:4`
        """
        return all(
            [
                self.start.library == self.end.library,
                self.start.book == self.end.book,
                self.start.chapter != self.end.chapter,
                (self.start.verse, self.end.verse) != (1, 999),
            ]
        )

    def is_verse(self) -> bool:
        """Determine if this range covers one verse only.

        Example:
            `Matthew 7:8`
        """
        return all(
            [
                self.start.library == self.end.library,
                self.start.book == self.end.book,
                self.start.chapter == self.end.chapter,
                self.start.verse == self.end.verse,
            ]
        )

    def is_verse_range(self) -> bool:
        """Determine if this range covers multiple verses within the same book
        and chapter.

        Example:
            `Matthew 7:8-12`
        """
        return all(
            [
                self.start.library == self.end.library,
                self.start.book == self.end.book,
                self.start.chapter == self.end.chapter,
                self.start.verse != self.end.verse,
            ]
        )


def range(start: Verse, end: Verse) -> Range:
    """A shorthand contructor for Range objects.

    Args:
        start: the `refspy.verse.Verse` at the start of the range.
        end: the `refspy.verse.Verse` at the end of the range.

    Note:
        To specify a single verse reference, supply a range with
        the same verse as the start and the end.
    """
    return Range(start=start, end=end)


def merge(ranges: List[Range], skip_sort: bool = False) -> List[Range]:
    """Merge overlapping ranges within a sorted list.

    This performs a sort before merging unless skip_sort=True.
    """
    if not ranges:
        return []
    sorted_ranges = ranges if skip_sort else sorted(ranges)
    new_ranges = []
    last_range = sorted_ranges[0]
    if sorted_ranges:
        for this_range in sorted_ranges[1:]:
            if last_range.overlaps(this_range):
                last_range = last_range.merge(this_range)
            else:
                new_ranges.append(last_range)
                last_range = this_range
    new_ranges.append(last_range)
    return new_ranges


def combine(ranges: List[Range], skip_merge: bool = False) -> List[Range]:
    """Join adjacent ranges within a sorted and merged list

    This performs a sort and merge before combining, unless skip_merge=True.
    """
    if not ranges:
        return []
    merged_ranges = ranges if skip_merge else merge(ranges)
    new_ranges = []
    last_range = merged_ranges[0]
    if merged_ranges:
        for this_range in merged_ranges[1:]:
            if last_range.adjoins(this_range):
                last_range = last_range.join(this_range)
            else:
                new_ranges.append(last_range)
                last_range = this_range
    new_ranges.append(last_range)
    return new_ranges


def book_range(
    library_id: Number, book_id: Number, book_end_id: Number | None = None
) -> Range:
    """
    Shorthand function for creating book references from `refspy.number.Number`
    values.
    """
    return range(
        verse(library_id, book_id, 1, 1),
        verse(library_id, book_end_id or book_id, 999, 999),
    )


def chapter_range(
    library_id: Number,
    book_id: Number,
    chapter_id: Number,
    chapter_end_id: Number | None = None,
) -> Range:
    """
    Shorthand function for creating chapter references from
    `refspy.number.Number` values.
    """
    return range(
        verse(library_id, book_id, chapter_id, 1),
        verse(library_id, book_id, chapter_end_id or chapter_id, 999),
    )


def verse_range(
    library_id: Number,
    book_id: Number,
    chapter_id: Number,
    verse_id: Number,
    verse_end_id: Number | None = None,
) -> Range:
    """
    Shorthand function for creating verse or range references from
    `refspy.number.Number` values.

    See `refspy.manager.Manager.bcv`
    """
    return range(
        verse(library_id, book_id, chapter_id, verse_id),
        verse(library_id, book_id, chapter_id, verse_end_id or verse_id),
    )



"""
Provide general utility functions for the Refspy package.
"""

import re
from typing import Any, List, Tuple

from refspy.number import Number


def parse_number(number_str: str) -> Number:
    """Remove non-digits and return an integer IF it falls between 1 and 999
    inclusive.

    Args:
        number_str: A string expected to contain an integer, among other
            characters.

    Raises:
        ValueError: if no digits in 1..999 range.

    Return:
        None if no digits appear in `number_str` or not in range.
    """
    digits = re.sub(r"\D", "", number_str)
    if digits != "":
        number = int(digits)
        if 1 <= number <= 999:
            return number
    raise ValueError(f"Invalid number string for parse_number: '{number_str}'")


def url_escape(name: str) -> str:
    """Turn a library or book name into a URL escaped string

    Simple URL encode for significant characters:

        * '1 Cor 3:4–5 -> '1%20Cor%203%3A4-5'
    """
    return name.replace(" ", "%20").replace(":", "%3A").replace("–", "-")


def url_param(name: str) -> str:
    """Turn a library or book name into a URL-friendly string.

    Lowercase and no spaces: '1 Cor 3:4-5' -> '1+cor+3.4-5'
    """
    return name.lower().replace(" ", "+").replace(":", ".").replace("–", "-")


def normalize_spacing(text: str) -> str:
    """Replace multiple spaces with single spaces.

    Args:
        text: A string expected to have irregular spacing.
    """
    return re.sub("\\s+", " ", text)


def pluralize(number: int, singular: str, plural: str = "") -> str:
    """
    Super simple pluralization.
    """
    if number == 1:
        return "%d %s" % (number, singular)
    else:
        if plural != "":
            return "%d %s" % (number, plural)
        else:
            return "%d %ss" % (number, singular)


def sequential_replace_tuples(text: str, tuples: List[Tuple[str, str]]):
    """Replace values, searching from the end of each previous replacement.

    This is for replacing matches, and ensures they are replaced in the
    same order they are matched.

    This function takes find-replace strings as a list of tuples. See
    `refspy.utils.sequential_replace`.

    Args:
        text: A string in which matches should be replaces
        tuples: A list of find-replace tuples, in order of appearance.
    """
    new_text = ""
    cursor = 0
    for find, replace in tuples:
        if find and replace:
            next_match = text.find(find, cursor)
            if next_match != -1:
                new_text += text[cursor:next_match]
                new_text += replace
                cursor = next_match + len(find)
    new_text += text[cursor:]
    return new_text


def sequential_replace(text: str, find: List[str], replace: List[str]) -> str:
    """Replace values, searching from the end of each previous replacement.

    This function takes find-replace strings as two separate lists. Unmatched
    entries will be ignored by zip(). See
    `refspy.utils.sequential_replace_tuples`.

    Args:
        text: A string in which matches should be replaces
        find: A list of values to be found, in order of appearance.
        replace: A list of replacement values.
    """
    return sequential_replace_tuples(text, list(zip(find, replace)))


def string_together(*args: Any) -> str:
    """Convert objects to strings and concatenate.

    Args:
        *args: A list of string-convertable values.
    """
    return "".join([str(arg) for arg in args])



"""Format Reference objects using Format objects."""

from typing import Dict, Tuple

from refspy.book import Book
from refspy.format import Format
from refspy.range import Range
from refspy.reference import Reference
from refspy.utils import string_together
from refspy.verse import Number, Verse


class Formatter:
    """
    Format an arbitrary reference, using supplied formatting options.

    References contain lists of `refspy.range.Range` objects. These can span
    arbitrary verses across multiple libraries, and can be arranged in any
    order. The formatter formats the individual ranges correctly and also
    indicates any changes between books and chapters.
    """

    def __init__(
        self,
        books: Dict[Tuple[Number, Number], Book],
        book_aliases: Dict[str, Tuple[Number, Number]],
    ) -> None:
        self.books = books
        self.book_aliases = book_aliases

    def format(
        self,
        reference: Reference,
        options: Format,
        if_invalid="[INVALID]",
    ) -> str:
        """
        Compare each range with the one before and decide whether we need to add
        the book name or chapter number to each range. This will work for
        sorted and unsorted references, but sorted references will be more compact.
        """
        out = ""
        last_verse = None
        for next_range in reference.ranges:
            if len(out) > 0:
                out += self.make_divider(next_range, last_verse, options)
            if next_range.is_book_range():
                out += self.make_book_range(next_range, last_verse, options)
            elif next_range.is_inter_book_range():
                out += self.make_inter_book_range(next_range, last_verse, options)
            elif next_range.is_book():
                out += self.make_book(next_range, last_verse, options)
            elif next_range.is_chapter_range():
                out += self.make_chapter_range(next_range, last_verse, options)
            elif next_range.is_inter_chapter_range():
                out += self.make_inter_chapter_range(next_range, last_verse, options)
            elif next_range.is_chapter():
                out += self.make_chapter(next_range, last_verse, options)
            elif next_range.is_verse_range():
                out += self.make_verse_range(next_range, last_verse, options)
            elif next_range.is_verse():
                out += self.make_verse(next_range, last_verse, options)
            else:
                out += if_invalid
            last_verse = next_range.start
        return out

    def make_divider(self, _: Range, last: Verse | None, options: Format) -> str:
        """Provide a separator between ranges.

        This will be a semicolon or comma.
        """
        if last is None:
            return ""
        if any(
            [
                _.start.library != last.library,
                _.start.book != last.book,
                _.start.chapter != last.chapter,
            ]
        ):
            return options.semicolon
        else:
            return options.comma

    def make_book_range(self, _: Range, last: Verse | None, options: Format) -> str:
        """Format a range between two whole books.

        See `refspy.range.Range.is_book_range` .

        Example:
            `Matthew-John`
        """
        start_book = self.books[_.start.library, _.start.book]
        start_book_name = getattr(start_book, options.property or "", "")
        end_book = self.books[_.end.library, _.end.book]
        end_book_name = getattr(end_book, options.property or "", "")
        return string_together(start_book_name, options.dash, end_book_name)

    def make_inter_book_range(
        self, _: Range, last: Verse | None, options: Format
    ) -> str:
        """Format a range that spans multiple books (or libraries).

        See `refspy.range.Range.is_inter_book_range` .

        Example:
            `Matt 15:15-John 15:15`
        """
        start_book = self.books[_.start.library, _.start.book]
        start_book_name = getattr(start_book, options.property or "", "")
        if _.start.verse == 1 and _.end.verse == 999:
            start_number = string_together(_.start.chapter)
        else:
            start_number = string_together(
                (_.start.chapter, options.colon, _.start.verse)
            )
        end_book = self.books[_.end.library, _.end.book]
        end_book_name = getattr(end_book, options.property or "", "")
        if _.end.verse == 1 and _.end.verse == 999:
            end_number = string_together(_.end.chapter)
        else:
            end_number = string_together(_.end.chapter, options.colon, _.end.verse)

        return options.space.join(
            [start_book_name, start_number, options.dash, end_book_name, end_number]
        )

    def make_book(self, _: Range, last: Verse | None, options: Format) -> str:
        """Format a range that spans multiple books (or libraries).

        See `refspy.range.Range.is_book` .

        Example:
            `Matthew`
        """
        start_book = self.books[_.start.library, _.start.book]
        start_book_name = getattr(start_book, options.property or "", "")
        return start_book_name

    def book_name_if_required(
        self, _: Range, last: Verse | None, options: Format
    ) -> str:
        """Format the book name (and a space) if the next range is in a
        different book.

        Example:
            `Matthew `
        """
        if options.number_only:
            return ""
        book_name = ""
        if last is None:
            book_name = self.make_book(_, last, options)
        elif _.start.library != last.library or _.start.book != last.book:
            book_name = self.make_book(_, last, options)
        if book_name != "":
            if options.book_only:
                return book_name
            else:
                return book_name + options.space
        return ""

    def make_chapter_range(self, _: Range, last: Verse | None, options: Format) -> str:
        """Format a range that spans whole chapters.

        Add the book name, if required.

        See `refspy.range.Range.is_chapter_range` .

        Example:
            `1-2` or `Matthew 1-2`
        """
        return string_together(
            self.book_name_if_required(_, last, options),
            _.start.chapter,
            options.dash,
            _.end.chapter,
        )

    def make_inter_chapter_range(
        self, _: Range, last: Verse | None, options: Format
    ) -> str:
        """Format a range that spans different chapters.

        Add the book name, if required.

        See `refspy.range.Range.is_inter_chapter_range` .

        Example:
            `1:1-2:2` or `Matthew 1:1-2:2`
        """
        return string_together(
            self.book_name_if_required(_, last, options),
            _.start.chapter,
            options.colon,
            _.start.verse,
            options.dash,
            _.end.chapter,
            options.colon,
            _.end.verse,
        )

    def make_chapter(self, _: Range, last: Verse | None, options: Format) -> str:
        """Format a range that spans different chapters.

        Add the book name, if required.

        See `refspy.range.Range.is_chapter`.

        Example:
            `1` or `Matthew 1`
        """
        return string_together(
            self.book_name_if_required(_, last, options), _.start.chapter
        )

    def chapter_number_if_required(
        self, _: Range, last: Verse | None, options: Format
    ) -> str:
        """Format the chapter number and colon if the next range is in a
        different book.

        Example:
            `1:`
        """
        if options.book_only:
            return ""
        book = self.books[_.start.library, _.start.book]
        if last is None:
            if book.chapters > 1:
                return string_together(_.start.chapter, options.colon)
        elif (
            _.start.library != last.library
            or _.start.book != last.book
            or _.start.chapter != last.chapter
        ):
            if book.chapters > 1:
                return string_together(_.start.chapter, options.colon)
        return ""

    def make_verse_range(self, _: Range, last: Verse | None, options: Format) -> str:
        """Format a range of different verses within a single chapter.

        Add the chapter number and book name, if required.

        See `refspy.range.Range.is_verse_range` .

        Example:
            `2-3` or `1:2-3` or `Matthew 1:2-3`

        Note:
            We abbreviate the second number if possible. See `abbreviate_range()`.
        """
        if options.book_only:
            return ""
        return string_together(
            self.book_name_if_required(_, last, options),
            self.chapter_number_if_required(_, last, options),
            "" if options.book_only else _.start.verse,
            "" if options.book_only else options.dash,
            "" if options.book_only else _.end.verse,
        )

    def make_verse(self, _: Range, last: Verse | None, options: Format) -> str:
        """Format a single verse number.

        Add the chapter number and book name, if required.

        See `refspy.range.Range.is_verse` .

        Example:
            `2` or `1:2` or `Matthew 1:2`
        """
        return string_together(
            self.book_name_if_required(_, last, options),
            self.chapter_number_if_required(_, last, options),
            "" if options.book_only else _.start.verse,
        )



"""Data object for a library of books.

Libraries have names, abbrevs (short names), and a list of the books that they
contain.

Attributes:
    id: 7
    name: '1 Corinthians'
    abbrev: '1 Cor'
    books: [],

See `refspy.libraries.en_US` for a complete example.
"""

from typing import List

from pydantic import BaseModel

from refspy.book import Book
from refspy.number import Number


class Library(BaseModel):
    id: Number
    name: str
    abbrev: str
    books: List[Book]



import pytest
from context import *

from refspy.indexers import (
    add_unique_book_alias,
    index_book_aliases,
    index_books,
    index_libraries,
)
from refspy.book import Book
from refspy.library import Library

bk_1 = Book(id=1, name="Book", abbrev="Bk", aliases=["vol"], chapters=10)
bk_2 = Book(id=2, name="Book", abbrev="Bk", aliases=[], chapters=10)
bk_3 = Book(id=3, name="3 Book", abbrev="3 Bk", aliases=[], chapters=1)

lib_1 = Library(id=1, name="Library", abbrev="Lib", books=[bk_1])
lib_2 = Library(id=2, name="Library", abbrev="Lib", books=[bk_1, bk_2, bk_3])


def test_book_uniqueness():
    index = dict()
    add_unique_book_alias(index, "Library", 1, 1)
    add_unique_book_alias(index, "Other Library", 1, 2)
    with pytest.raises(ValueError):
        add_unique_book_alias(index, "Library", 1, 3)


def test_index_books():
    index = index_books([lib_2])
    assert len(index) == 3
    assert index[2, 1] == bk_1
    assert index[2, 2] == bk_2


def test_index_book_aliases():
    index = index_book_aliases([lib_1])
    assert len(index) == 3
    assert index["Book"] == (1, 1)
    assert index["Bk"] == (1, 1)


def test_non_strict_index_book_aliases_with_params():
    index = index_book_aliases([lib_2], url_param_names=True, strict=False)
    assert len(index) == 5
    assert index["book"] == (2, 2)
    assert index["bk"] == (2, 2)
    assert index["3+book"] == (2, 3)
    assert index["3+bk"] == (2, 3)


def test_index_libraries():
    index = index_libraries([lib_1])
    assert len(index) == 1
    assert index[1] == lib_1



from context import *

from refspy.languages.english import ENGLISH
from refspy.libraries.en_US import OT, NT
from refspy.manager import Manager
from refspy.range import range, verse_range
from refspy.reference import reference, verse_reference
from refspy.verse import verse

__ = Manager(libraries=[OT, NT], language=ENGLISH)


def test_number_ranges():
    refs = __.find_references("Romans 1:1–4, 6, 7-9")
    assert refs[0][1] == reference(
        verse_range(NT.id, 6, 1, 1, 4),
        verse_range(NT.id, 6, 1, 6),
        verse_range(NT.id, 6, 1, 7, 9),
    )


def test_multi_chapter_ranges():
    _, ref = __.first_reference("Rom 1:1-2:4")
    assert ref == reference(range(verse(NT.id, 6, 1, 1), verse(NT.id, 6, 2, 4)))


def test_one_chapter_books():
    refs = __.find_references("Philemon 3-6, v.15")
    assert refs[0][1] == reference(verse_range(NT.id, 18, 1, 3, 6))
    assert refs[1][1] == reference(verse_range(NT.id, 18, 1, 15))


def test_verse_markers():
    refs = __.find_references("Romans 1:1, v.9, vv.11-12")
    assert refs[0][1] == verse_reference(NT.id, 6, 1, 1)
    assert refs[1][1] == verse_reference(NT.id, 6, 1, 9)
    assert refs[2][1] == verse_reference(NT.id, 6, 1, 11, 12)


def test_ambiguous_verses():
    refs = __.find_references("v.5, v.6-7, 6:5, 8:6-7")
    assert len(refs) == 0


def test_ambiguous_verse_return_none():
    refs = __.find_references("v.5, v.6-7, 6:5, 8:6-7", include_nones=True)
    assert len(refs) == 4


def test_ambiguous_context():
    refs = __.find_references("v.5", include_nones=True)
    assert len(refs) == 1
    refs = __.find_references("8:5", include_nones=True)
    assert len(refs) == 1
    refs = __.find_references("Romans, v.5", include_nones=True)
    assert len(refs) == 1


def test_ambiguous_context_returns_none():
    refs = __.find_references("Romans, v.5", include_nones=True)
    assert len(refs) == 1
    assert refs[0][1] == None


def test_books_with_number_prefixes():
    refs = __.find_references("Romans 1:1, 2, 3 John 3")
    #             MATCH THIS:  ^^^^^^^^^^^^^  ^^^^^^^^
    #               NOT THIS:  ^^^^^^^^^^^^^^^^ ^^^^^^
    assert refs[0][1] == reference(
        verse_range(NT.id, 6, 1, 1), verse_range(NT.id, 6, 1, 2)
    )
    assert refs[1][1] == verse_reference(NT.id, 25, 1, 3)


def test_substitute_number_prefixes():
    refs = __.find_references("I Cor 2:3; First Cor 2:3; 1st Cor 2:3")
    assert refs[0][1] == refs[1][1] == refs[2][1]
    refs = __.find_references("II Thess 2:3, Second Thess 2:3, 2nd Thess 2:3")
    assert refs[0][1] == refs[1][1] == refs[2][1]
    refs = __.find_references("III John 2; Third John 2; 3rd John 2")
    assert refs[0][1] == refs[1][1] == refs[2][1]


def test_backward_ranges_and_abbreviations():
    refs = __.find_references("Romans 1:4–1, Rom 1:776-77, Rom 1:13-4")
    # No match for first reference
    assert len(refs) == 2
    assert refs[0][1] == verse_reference(NT.id, 6, 1, 776, 777)
    assert refs[1][1] == verse_reference(NT.id, 6, 1, 13, 14)


def test_partial_verses():
    _, ref = __.first_reference("Rom 1:1a-2:4b")
    #                                   ^    ^  <-- Ignore letters
    assert ref == reference(range(verse(NT.id, 6, 1, 1), verse(NT.id, 6, 2, 4)))


def test_brackets():
    refs = __.find_references("Romans (John 1:1–4 (1 Cor 5:24 ()) 3:16) 1:16-17")
    #                    MATCH:  xxxxxx  ^^^^^^^^^^  ^^^^^^^^^^     ^^^^  ^^^^^^^
    #                            ^ Don't match this, but use for context: ^^^^^^^
    assert len(refs) == 4
    assert refs[0][1] == verse_reference(NT.id, 4, 1, 1, 4)
    assert refs[1][1] == verse_reference(NT.id, 7, 5, 24)
    assert refs[2][1] == verse_reference(NT.id, 4, 3, 16)
    assert refs[3][1] == verse_reference(NT.id, 6, 1, 16, 17)


def test_malformed_brackets():
    refs = __.find_references("Romans )John 1:1–4 (1 Cor 5:24 (() 3:16) 1:16-17")
    #                            ^^^ Ignore          ^^^ 1 Corinthians
    assert refs[0][1] == verse_reference(NT.id, 4, 1, 1, 4)
    assert refs[1][1] == verse_reference(NT.id, 7, 5, 24)
    assert refs[2][1] == verse_reference(NT.id, 7, 3, 16)
    assert refs[3][1] == verse_reference(NT.id, 7, 1, 16, 17)



"""Indexing functions for libraries, books, and book aliases."""

from typing import Dict, List, Tuple
from refspy.book import Book
from refspy.library import Library
from refspy.utils import url_param
from refspy.verse import Number

PARAM_CHAR_LIMIT = 8
"""The maximum alias length to convert to URL parameter names."""


def index_libraries(libraries: List[Library]) -> Dict[Number, Library]:
    """Create a lookup dict for libraries by library.id"""
    index = dict()
    for library in libraries:
        index[library.id] = library
    return index


def index_books(libraries: List[Library]) -> Dict[Tuple[Number, Number], Book]:
    """Create a lookup dict for books by (library.id, book.id)"""
    index = dict()
    for library in libraries:
        for book in library.books:
            index[library.id, book.id] = book
    return index


def add_unique_book_alias(
    index: Dict[str, Tuple[Number, Number]],
    alias: str,
    library_id: Number,
    book_id: Number,
    strict: bool = True,
):
    """Add an index entry for a library and book if none exists

    Modifies the passed index dict.

    Args:
        strict: raise an error in the case of a collision between
            alias values.
    """
    if alias in index and strict:
        raise ValueError(f"Book alias '{alias}' is not unique.")
    if alias == "":
        raise ValueError(f"Book alias is '' for ({library_id},{book_id})")
    else:
        index[alias] = (library_id, book_id)


def index_book_aliases(
    libraries: List[Library], url_param_names: bool = False, strict: bool = True
) -> Dict[str, Tuple[Number, Number]]:
    """Create a lookup table for library and book IDs by book aliases.

    Args:
        libraries: A list of `refspy.library.Library` objects to be indexed.
        url_param_names: Convert aliases to url parameters if they are less
            than `PARAM_CHAR_LIMIT` in length.
        strict: Throw a ValueError if any two aliases are the same, or any
            alias is an empty string.

    Note:
        Because books like Ruth or Mark have book.abbrev == book.name, we only
        index the abbrev if it differs from the name. These books will not raise
        a ValueError when strict=True.
    """
    index = dict()
    for library in libraries:
        for book in library.books:
            if not url_param_names or len(book.name) < PARAM_CHAR_LIMIT:
                add_unique_book_alias(
                    index,
                    url_param(book.name) if url_param_names else book.name,
                    library.id,
                    book.id,
                    strict,
                )
            if book.name != book.abbrev:
                if not url_param_names or len(book.abbrev) < PARAM_CHAR_LIMIT:
                    add_unique_book_alias(
                        index,
                        url_param(book.abbrev) if url_param_names else book.abbrev,
                        library.id,
                        book.id,
                        strict,
                    )
            for alias in book.aliases:
                if not url_param_names or len(alias) < PARAM_CHAR_LIMIT:
                    add_unique_book_alias(
                        index,
                        url_param(alias) if url_param_names else alias,
                        library.id,
                        book.id,
                        strict,
                    )
    return index



import re

import pytest
from context import *

from refspy.indexers import index_book_aliases, index_books
from refspy.languages.english import ENGLISH
from refspy.matcher import (
    CHAPTER_RANGE_CAPTURE,
    CHAPTER_VERSES_CAPTURE,
    COLON,
    DASH,
    LIST,
    NUMBER,
    RANGE,
    RANGE_OR_NUMBER_COMPILED,
    SPACE,
    Matcher,
    infer_abbreviation,
    make_chapter_range,
    make_chapter_verses,
    make_number_ranges,
    match_chapter_range,
    match_chapter_verses,
    match_number_ranges,
)
from refspy.range import range, verse_range
from refspy.reference import (
    book_reference,
    chapter_reference,
    reference,
    verse_reference,
)
from refspy.verse import verse
from tests.data import TEST_LIBRARY

books = index_books([TEST_LIBRARY])
book_aliases = index_book_aliases([TEST_LIBRARY])
matcher = Matcher(books, book_aliases, ENGLISH)


def test_regexp_building_blocks():
    assert re.findall(COLON, ":") == [":"]

    assert re.findall(DASH, "-") == ["-"]
    assert re.findall(DASH, "–") == ["–"]

    assert re.findall(NUMBER, "0") == ["0"]
    assert re.findall(NUMBER, "1") == ["1"]
    assert re.findall(NUMBER, "1000") == ["1000"]  # <-- numbers of any length

    assert re.findall(SPACE, " ") == [" "]
    assert re.findall(SPACE, "\n") == ["\n"]
    assert re.findall(SPACE, "\r") == ["\r"]
    assert re.findall(SPACE, "\t") == ["\t"]

    assert re.findall(RANGE, "1-2") == ["1-2"]
    assert re.findall(RANGE, "0-999") == ["0-999"]
    assert re.findall(RANGE, "1-1000") == ["1-1000"]
    assert re.findall(RANGE, "1-999") == ["1-999"]


def test_number_list_regexp():
    match = re.findall(LIST, "1-3,4:5,7-9")
    assert match == ["1-3,4", "5,7-9"]
    match = re.findall(LIST, "1-3,4,5, 7-9")
    assert match == ["1-3,4,5, 7-9"]


def test_number_list_generated_regexp():
    number_list = matcher.build_number_list_regexp()
    match = re.findall(number_list, "Big Book 1:1, 1, 1 Book 2, 3, 4:5")
    assert match == ["1", "1, 1", "1", "2, 3", "4", "5"]
    #                      ^^^^ not 1, 1, 1


def test_range_or_number_regexp():
    match = RANGE_OR_NUMBER_COMPILED.findall("1,2-3, 4,   5-6")
    assert match == ["1", "2-3", "4", "5-6"]


def test_chapter_range_regexp():
    match = CHAPTER_RANGE_CAPTURE.search(" 1:2-3:4")
    assert match is not None
    assert match.group(1) == "1"
    assert match.group(2) == "2"
    assert match.group(3) == "3"
    assert match.group(4) == "4"


def test_chapter_verses_regexp():
    match = CHAPTER_VERSES_CAPTURE.search(" 1:2,3-4")
    assert match is not None
    assert match.group(1) == "1"
    assert match.group(2) == "2,3-4"


def test_name_regexp():
    regexp = matcher.build_reference_regexp()
    assert re.escape("Big") + "\\s+" + re.escape("Book") in regexp
    assert re.escape("Small") + "\\s+" + re.escape("Book") in regexp


def test_match_brackets():
    text = "Big Book 1:2-5 (and 34:6 (and 7)) are more interesting than Small Book 3-6."
    matches = matcher.brackets_regexp.findall(text)
    assert len(matches) == 4
    assert matches[0] == "("
    assert matches[1] == "("
    assert matches[2] == ")"
    assert matches[3] == ")"


def test_match_names():
    assert matcher.name_regexp.findall("Big Book") == [("Big Book", "Big Book", "", "")]
    assert matcher.name_regexp.findall("Big") == [("Big", "Big", "", "")]
    # assert matcher.name_regexp.findall("big") == [("big", "big", "", "")]
    assert matcher.name_regexp.findall("Small Book") == [
        ("Small Book", "Small Book", "", "")
    ]
    assert matcher.name_regexp.findall("Small") == [("Small", "Small", "", "")]
    # assert matcher.name_regexp.findall("small") == [("small", "small", "", "")]


def test_match_names_and_numbers():
    assert matcher.name_regexp.findall("Big Book 1") == [
        ("Big Book 1", "Big Book", " 1", "")
    ]
    assert matcher.name_regexp.findall("Big Book 1-2") == [
        ("Big Book 1-2", "Big Book", " 1-2", "")
    ]
    assert matcher.name_regexp.findall("Big Book 1-2,4,6") == [
        ("Big Book 1-2,4,6", "Big Book", " 1-2,4,6", "")
    ]
    assert matcher.name_regexp.findall("Big Book 1-2,4,6") == [
        ("Big Book 1-2,4,6", "Big Book", " 1-2,4,6", "")
    ]


def test_malformed_refs():
    matches = match_number_ranges("1-2,4-3")
    assert matches[0] == "1-2"
    assert matches[1] == "4-3"


def test_commas_separation():
    text = "Big Book 1:1, 2, 1 Bk 3, Small Book 4"
    #                     ^^^^ Negative lookaheads will prevent confusion here.
    matches = matcher.name_regexp.findall(text)
    assert len(matches) == 3
    assert matches[0] == ("Big Book 1:1, 2", "Big Book", " 1:1, 2", "")
    assert matches[1] == ("1 Bk 3", "1 Bk", " 3", "")
    assert matches[2] == ("Small Book 4", "Small Book", " 4", "")


def test_match_names_in_context():
    text = "Big Book 1:2-5 and 34:6,7 are more interesting than Small Book 3-6."
    matches = matcher.name_regexp.findall(text)
    assert len(matches) == 3
    assert matches[0] == ("Big Book 1:2-5", "Big Book", " 1:2-5", "")
    assert matches[1] == ("", "", "", "34:6,7")
    assert matches[2] == ("Small Book 3-6", "Small Book", " 3-6", "")


def test_match_number_ranges_numeric():
    last_range = verse_range(1, 1, 1, 1)
    matches = match_number_ranges("1,4-5,7–8, 9")
    reference = make_number_ranges(last_range, matches)
    assert reference is not None
    assert reference.ranges[0] == range(verse(1, 1, 1, 1), verse(1, 1, 1, 1))
    assert reference.ranges[1] == range(verse(1, 1, 1, 4), verse(1, 1, 1, 5))
    assert reference.ranges[2] == range(verse(1, 1, 1, 7), verse(1, 1, 1, 8))
    assert reference.ranges[3] == range(verse(1, 1, 1, 9), verse(1, 1, 1, 9))


def test_match_number_ranges_prefixed_v():
    last_range = verse_range(1, 1, 1, 1)
    matches = match_number_ranges("v.1,4-5")
    assert matches is not None
    reference = make_number_ranges(last_range, matches)
    assert reference is not None
    assert reference.ranges[0] == range(verse(1, 1, 1, 1), verse(1, 1, 1, 1))
    assert reference.ranges[1] == range(verse(1, 1, 1, 4), verse(1, 1, 1, 5))


def test_match_number_ranges_prefixed_vv():
    last_range = verse_range(1, 1, 1, 1)
    matches = match_number_ranges("vv.1,4-5")
    assert matches is not None
    reference = make_number_ranges(last_range, matches)
    assert reference is not None
    assert reference.ranges[0] == range(verse(1, 1, 1, 1), verse(1, 1, 1, 1))
    assert reference.ranges[1] == range(verse(1, 1, 1, 4), verse(1, 1, 1, 5))


def test_match_chapter_range():
    last_range = verse_range(1, 1, 1, 1)
    matches = match_chapter_range("1:4-2:3")
    assert matches is not None
    reference = make_chapter_range(last_range, matches)
    assert reference is not None
    assert reference.ranges == [range(verse(1, 1, 1, 4), verse(1, 1, 2, 3))]
    matches = match_chapter_range("1:4–2:3")
    assert matches is not None
    reference = make_chapter_range(last_range, matches)
    assert reference is not None
    assert reference.ranges == [range(verse(1, 1, 1, 4), verse(1, 1, 2, 3))]


def test_match_chapter_verses():
    last_range = verse_range(1, 1, 1, 1)
    matches = match_chapter_verses("1:4,8-9")
    assert matches is not None
    reference = make_chapter_verses(last_range, matches)
    assert reference is not None
    assert reference.ranges == [
        range(verse(1, 1, 1, 4), verse(1, 1, 1, 4)),
        range(verse(1, 1, 1, 8), verse(1, 1, 1, 9)),
    ]


def test_find_references():
    sample_text = "Big Book 1:2-5 and 34:6,7 are more interesting than Small Book 3-6."
    __ = matcher.generate_references(sample_text)
    text, ref = next(__)
    assert text == "Big Book 1:2-5"
    assert ref == reference(range(verse(1, 2, 1, 2), verse(1, 2, 1, 5)))
    text, ref = next(__)
    assert text == "34:6,7"
    assert ref == reference(
        range(verse(1, 2, 34, 6), verse(1, 2, 34, 6)),
        range(verse(1, 2, 34, 7), verse(1, 2, 34, 7)),
    )
    text, ref = next(__)
    assert text == "Small Book 3-6"
    assert ref == reference(range(verse(1, 3, 1, 3), verse(1, 3, 1, 6)))
    with pytest.raises(StopIteration):
        assert next(__) is None


def test_vv_references():
    sample_text = "Big Book 5 is useful, esp. vv.3-4 and v.8."
    __ = matcher.generate_references(sample_text)
    text, ref = next(__)
    assert text == "Big Book 5"
    assert ref == reference(range(verse(1, 2, 5, 1), verse(1, 2, 5, 999)))
    text, ref = next(__)
    assert text == "vv.3-4"
    assert ref == reference(range(verse(1, 2, 5, 3), verse(1, 2, 5, 4)))
    text, ref = next(__)
    assert text == "v.8"
    assert ref == reference(range(verse(1, 2, 5, 8), verse(1, 2, 5, 8)))
    with pytest.raises(StopIteration):
        assert next(__) is None


def test_include_books():
    sample_text = "Big Book or Small Book?"
    __ = matcher.generate_references(sample_text, include_books=True)
    text, ref = next(__)
    assert text == "Big Book"
    assert ref == book_reference(1, 2)
    text, ref = next(__)
    assert text == "Small Book"
    assert ref == book_reference(1, 3)
    # ^^ The final reference should still be to Big Book.
    with pytest.raises(StopIteration):
        assert next(__) is None


def test_single_parentheses():
    sample_text = "Big Book 1:2-5 (cf. Small Book 34) is more interesting than vv.3-6."
    # ^^ The final reference should still be to Big Book (book ID 2).
    __ = matcher.generate_references(sample_text)
    text, ref = next(__)
    assert text == "Big Book 1:2-5"
    assert ref == reference(range(verse(1, 2, 1, 2), verse(1, 2, 1, 5)))
    text, ref = next(__)
    assert text == "Small Book 34"
    assert ref == reference(range(verse(1, 3, 1, 34), verse(1, 3, 1, 34)))
    text, ref = next(__)
    assert text == "vv.3-6"
    assert ref == reference(range(verse(1, 2, 1, 3), verse(1, 2, 1, 6)))
    with pytest.raises(StopIteration):
        assert next(__) is None


def test_number_prefixes():
    sample_text = "1 Book 1:1, First Book 1:1, 1st Book 1:1, I Book 1:1"
    first_reference = verse_reference(1, 4, 1, 1)
    __ = matcher.generate_references(sample_text)
    text, ref = next(__)
    assert text == "1 Book 1:1"
    assert ref == first_reference
    text, ref = next(__)
    assert text == "First Book 1:1"
    assert ref == first_reference
    text, ref = next(__)
    assert text == "1st Book 1:1"
    assert ref == first_reference
    text, ref = next(__)
    assert text == "I Book 1:1"
    assert ref == first_reference
    with pytest.raises(StopIteration):
        assert next(__) is None


def test_line_wrapping():
    text = """
        This is a wrapped reference to 1 
        Book 5.
        """
    __ = matcher.generate_references(text)
    text, ref = next(__)
    assert (
        text
        == """1 
        Book 5"""
    )
    assert ref == chapter_reference(1, 4, 5)


def test_infer_abbreviation():
    assert infer_abbreviation("124", "24") == "124"
    assert infer_abbreviation("12", "4") == "14"
    assert infer_abbreviation("24", "23") is None
    assert infer_abbreviation("4", "3") is None



"""Data type for Numbers

Library, book, chapter, and verse numbers are integers in the range 1 to 999.
"""

from typing import Annotated
from annotated_types import Ge, Le  # >=, <=

Number = Annotated[int, Ge(1), Le(999)]



"""Data object for a verse.

Attributes:
    library: 1
    book: 2
    chapter: 3
    verse: 4
"""

from typing import Tuple, Self

from pydantic import BaseModel

from refspy.index import Index
from refspy.number import Number


VerseTuple = Tuple[Number, Number, Number, Number]


class Verse(BaseModel):
    """
    Library, Book, Chapter, Verse
    """

    library: Number
    book: Number
    chapter: Number
    verse: Number

    def tuple(self) -> VerseTuple:
        """A sortable tuple for comparison operations."""
        return (self.library, self.book, self.chapter, self.verse)

    def __lt__(self, other: Self) -> bool:
        """Compare verses by comparing tuples."""
        return self.tuple() < other.tuple()

    def __le__(self, other: Self) -> bool:
        """Compare verses by comparing tuples."""
        return self.tuple() <= other.tuple()

    def __gt__(self, other: Self) -> bool:
        """Compare verses by comparing tuples."""
        return self.tuple() > other.tuple()

    def __ge__(self, other: Self) -> bool:
        """Compare verses by comparing tuples."""
        return self.tuple() >= other.tuple()

    @classmethod
    def from_index(cls, index: Index) -> Self:
        """
        Convert a verse's index number to its tuple form.

        Example:
            ```
            verse = Verse.from_index(1002003004)
            ```
        """
        lbc, v = divmod(index, 1000)
        lb, c = divmod(lbc, 1000)
        l, b = divmod(lb, 1000)
        return cls(library=l, book=b, chapter=c, verse=v)

    def index(self) -> Index:
        """Create a `refspy.index.Index` number representing this verse.

        Example:
            `verse(1, 2, 3, 4)` becomes `1002003004`
        """
        return (
            (self.library * 1000 + self.book) * 1000 + self.chapter
        ) * 1000 + self.verse


def verse(library: Number, book: Number, chapter: Number, verse: Number) -> Verse:
    """A shorthand constructor for verses."""
    return Verse(library=library, book=book, chapter=chapter, verse=verse)



from pydantic import ValidationError
import pytest

from refspy.verse import verse, Verse


def test_shorthand():
    _ = verse(1, 2, 3, 4)
    assert _.library == 1
    assert _.book == 2
    assert _.chapter == 3
    assert _.verse == 4


def test_from_index():
    _ = Verse.from_index(1002003004)
    assert _.library == 1
    assert _.book == 2
    assert _.chapter == 3
    assert _.verse == 4


def test_keyword_args():
    _ = Verse(library=1, book=2, chapter=3, verse=4)
    assert _.library == 1
    assert _.book == 2
    assert _.chapter == 3
    assert _.verse == 4


def test_args_out_of_number_range():
    with pytest.raises(ValidationError):
        _ = verse(0, 2, 3, 4)
    with pytest.raises(ValidationError):
        _ = verse(1000, 2, 3, 4)
    with pytest.raises(ValidationError):
        _ = verse(1, 0, 3, 4)
    with pytest.raises(ValidationError):
        _ = verse(1, 1000, 3, 4)
    with pytest.raises(ValidationError):
        _ = verse(1, 2, 0, 4)
    with pytest.raises(ValidationError):
        _ = verse(1, 2, 1000, 4)
    with pytest.raises(ValidationError):
        _ = verse(1, 2, 3, 0)
    with pytest.raises(ValidationError):
        _ = verse(1, 2, 3, 1000)


def test_tuples():
    _ = verse(1, 2, 3, 4)
    assert _.tuple() == (1, 2, 3, 4)


def test_index_numbers():
    _ = verse(1, 2, 3, 4)
    assert _.index() == 1002003004



"""A facade for all major interface functions.

See `refspy.refspy()` for a useful helper function.
"""

import re
from typing import Dict, Generator, List, Tuple

from pydantic import TypeAdapter
from refspy.book import Book
from refspy.format import (
    ABBREV_BOOK_FORMAT,
    ABBREV_NAME_FORMAT,
    BOOK_FORMAT,
    NAME_FORMAT,
    NUMBER_FORMAT,
)
from refspy.formatter import Formatter
from refspy.indexers import (
    index_book_aliases,
    index_books,
    index_libraries,
)
from refspy.language import Language
from refspy.library import Library
from refspy.matcher import Matcher
from refspy.navigator import Navigator
from refspy.range import combine, merge, range
from refspy.reference import (
    Reference,
    book_reference,
    chapter_reference,
    reference,
    verse_reference,
)
from refspy.utils import pluralize, url_param, url_escape
from refspy.verse import Number, verse


class Manager:
    """
    The manager object integrates the main library features into a single
    convenient facade.
    """

    def __init__(self, libraries: List[Library], language: Language):
        self.libraries: Dict[Number, Library] = index_libraries(libraries)
        """A lookup dictionary for Libraries by library.id """

        self.books: Dict[Tuple[Number, Number], Book] = index_books(libraries)
        """A lookup dictionary for Books by (library.id, book.id)"""

        self.book_aliases: Dict[str, Tuple[Number, Number]] = index_book_aliases(
            libraries
        )
        """A lookup dictionary for (library.id, book.id) by book alias strings."""

        self.language: Language = language
        """Language-specific program data."""

        self.matcher: Matcher = Matcher(self.books, self.book_aliases, self.language)
        """Delegate reference matching tasks."""

        self.formatter: Formatter = Formatter(self.books, self.book_aliases)
        """Delegate formatting tasks."""

        self.navigator: Navigator = Navigator(self.books, self.book_aliases)
        """Delegate navigation tasks."""

    # -----------------------------------
    # Index and summary functions
    # -----------------------------------

    def make_index(
        self, references: List[Reference], pattern: str | None = None
    ) -> str:
        """Write a one-line index of references.

        Args:
            pattern: a formatting string suitable for links, see
                `refspy.manager.Manager.template()`.

        Note:
            If no template pattern is provided, reference formatting
            will default to `refspy.manager.Manager.abbrev_name()`
        """
        index = []
        collation = self.collate(
            sorted([ref for ref in references if ref and not ref.is_book()])
        )
        for _, book_collation in collation:
            for _, reference_list in book_collation:
                sorted_ref = self.sort_references(reference_list)
                if pattern:
                    index.append(self.template(sorted_ref, pattern))
                else:
                    index.append(self.abbrev_name(sorted_ref))
        return "; ".join(index)

    def make_summary(
        self, references: List[Reference], pattern: str | None = None
    ) -> str:
        """Write a one-line summary, like for index, but merge and join ranges.

        Args:
            pattern: a formatting string suitable for links, see
                `refspy.manager.Manager.template()`.

        Note:
            If no template pattern is provided, reference formatting
            will default to `refspy.manager.Manager.abbrev_name()`.
        """
        summary = []
        count_nones = len([ref for ref in references if ref is None])
        if count_nones > 0:
            error_msg = f'; <span class="wiki-error">{pluralize(count_nones, "bad reference")}</span>'
        else:
            error_msg = ""
        collation = self.collate(
            sorted([ref for ref in references if ref and not ref.is_book()])
        )
        for _, book_collation in collation:
            for _, reference_list in book_collation:
                compact_ref = self.combine_references(reference_list)
                if pattern:
                    summary.append(self.template(compact_ref, pattern))
                else:
                    summary.append(self.abbrev_name(compact_ref))
        return "; ".join(summary) + error_msg

    def make_hotspots(
        self,
        references: List[Reference],
        max_chapters: int = 7,
        min_references: int = 2,
    ) -> List[Tuple[str, int]] | None:
        """
        Find the most referenced chapters in a set of references.

        Args:
            max_chapters: The maximum number of chapter hotspots to return.
            min_references: The minimal references per chapter that qualifies as a hotspot.
        """
        totals = dict()
        if not references:
            return None
        for _ in self.sort_references(references).ranges:
            for tuple in [
                (_.start.library, _.start.book, _.start.chapter),
                (_.end.library, _.end.book, _.end.chapter),
            ]:
                if tuple in totals:
                    totals[tuple] += 1
                else:
                    totals[tuple] = 1
        hotspots = {
            self.abbrev_name(chapter_reference(*tuple)): int(total / 2)
            for tuple, total in totals.items()
            if total / 2 >= min_references
        }
        hotspots_desc = sorted(hotspots.items(), key=lambda item: item[1], reverse=True)
        return hotspots_desc[:max_chapters]

    def make_hotspots_text(
        self,
        references: List[Reference],
        max_chapters: int = 7,
        min_references: int = 2,
    ) -> str | None:
        """
        Return most referenced chapters as a text string: "Rom 3, 1 Cor 6" in
        descending order of frequency.

        Args:
            max_chapters: The maximum number of chapter hotspots to return.
            min_references: The minimal references per chapter that qualifies as a hotspot.
        """
        if hotspots := self.make_hotspots(
            references, max_chapters=max_chapters, min_references=min_references
        ):
            return ", ".join([chapter for chapter, _ in hotspots])
        else:
            return None

    # -----------------------------------
    # Merging functions
    # -----------------------------------

    def sort_references(self, references: List[Reference]) -> Reference:
        """For a list of references, make a single reference containing their sorted ranges."""

        ranges = []
        for ref in references:
            for rng in ref.ranges:
                ranges.append(rng)
        return reference(*sorted(ranges))

    def merge_references(self, references: List[Reference]) -> Reference:
        """For a list of references, merge their ranges into a new reference.

        Merging means combining ranges that overlap.
        """
        ranges = []
        for ref in references:
            ranges.extend(ref.ranges)
        new_ref = reference(*merge(ranges))
        return new_ref

    def combine_references(self, references: List[Reference]) -> Reference:
        """For a list of references, combine their ranges into a new reference.

        Combining means sorting ranges, merging overlaps, and joining adjacent
        records.
        """
        ranges = []
        for ref in references:
            ranges.extend(ref.ranges)
        new_ref = reference(*combine(ranges))
        return new_ref

    # -----------------------------------
    # Iteration functions
    # -----------------------------------

    def collate(
        self, references: List[Reference]
    ) -> List[Tuple[Library, List[Tuple[Book, List[Reference]]]]]:
        """
        A collation groups single-book references by library and book,
        providing Library and Book objects for iteration. Multi-book references
        are ignored.
        """
        library_list = list()
        for library_id, books_dict in self.collate_by_id(references).items():
            book_list = list()
            for book_id, references in books_dict.items():
                book_list.append(tuple([self.books[library_id, book_id], references]))
            library_list.append(tuple([self.libraries[library_id], book_list]))
        return library_list

    def collate_by_id(
        self, references: List[Reference]
    ) -> Dict[Number, Dict[Number, List[Reference]]]:
        """
        A collation groups single-book references by library and book IDs.
        Multi-book references are ignored.
        """
        collation = dict()
        for ref in [_ for _ in references if _.count_books() == 1]:
            v1 = ref.ranges[0].start
            if v1.library not in collation:
                collation[v1.library] = dict()
            if v1.book not in collation[v1.library]:
                collation[v1.library][v1.book] = list()
            collation[v1.library][v1.book].append(ref)
        return collation

    # -----------------------------------
    # Matching functions
    # -----------------------------------

    def first_reference(self, text: str) -> Tuple[str | None, Reference | None]:
        """
        Return the first tuple of (match_str, reference) found by
        `refspy.manager.Manager.generate_references()`
        """
        generator = self.matcher.generate_references(text)
        return next(generator, (None, None))

    def find_references(
        self, text: str, include_books: bool = False, include_nones: bool = False
    ) -> List[Tuple[str, Reference | None]]:
        """
        Return a list of tuples of (match_str, reference) found by
        `refspy.manager.Manager.generate_references()`
        """
        generator = self.matcher.generate_references(text, include_books, include_nones)
        return list(generator)

    def generate_references(
        self, text: str, include_books: bool = False, include_nones: bool = False
    ) -> Generator[Tuple[str, Reference | None], None, None]:
        """
        Generate tuples of (match_str, reference) for provided text.

        Task delegated to `refspy.matcher.Matcher.generate_references()`.

        Args:
            text: a string to search for references
            include_books: Whether to yield book names without reference numbers
            include_nones: Whether to yield (match_str, None) for malformed
                references.

        Yield:
            A tuple of `(match_str, reference)` for each valid reference.
        """
        yield from self.matcher.generate_references(text, include_books, include_nones)

    # -----------------------------------
    # Reference creator functions
    # -----------------------------------

    def bcv(
        self,
        alias: str,
        c: Number | None = None,
        v: Number | None = None,
        v_end: Number | None = None,
    ) -> Reference:
        """Construct a reference from a book alias and chapter/verse numbers.

        Omitting optional arguments will construct a reference to a whole book
        or chapter.

        Args:
            v_end: constructs a verse range from `v` to `v_end`.

        Raises:
            ValueError: If the book_alias is missing, or the number values are
                out of range.
        """
        if alias not in self.book_aliases:
            raise ValueError(f'Book alias "{alias}" not found.')
        library_id, book_id = self.book_aliases[alias]
        ta = TypeAdapter(Number)
        if c is None:
            return book_reference(library_id, book_id)
        if v is None:
            return chapter_reference(library_id, book_id, ta.validate_python(c))

        if not v_end:
            return verse_reference(
                library_id,
                book_id,
                ta.validate_python(c),
                ta.validate_python(v),
            )
        else:
            return reference(
                range(
                    verse(
                        library_id,
                        book_id,
                        ta.validate_python(c),
                        ta.validate_python(v),
                    ),
                    verse(
                        library_id,
                        book_id,
                        ta.validate_python(c),
                        ta.validate_python(v_end),
                    ),
                )
            )

    def bcr(
        self, alias: str, c: Number, v_ranges: List[Number | Tuple[Number, Number]]
    ) -> Reference:
        if alias not in self.book_aliases:
            raise ValueError(f'Book alias "{alias}" not found.')
        library_id, book_id = self.book_aliases[alias]
        ranges = []
        for val in v_ranges:
            if isinstance(val, tuple):
                v_start, v_end = val
                ranges.append(
                    range(
                        verse(library_id, book_id, c, v_start),
                        verse(library_id, book_id, c, v_end),
                    )
                )
            elif isinstance(val, int):
                v_start = v_end = val
                ranges.append(
                    range(
                        verse(library_id, book_id, c, v_start),
                        verse(library_id, book_id, c, v_end),
                    )
                )
        return reference(*ranges)

    def r(self, text: str) -> Reference | None:
        """Return the first matching reference in the given text.

        Book names and malformed references are not matched.
        """
        _, reference = self.first_reference(text)
        return reference

    # -----------------------------------
    # Navigation functions
    # -----------------------------------

    def next_chapter(self, ref: Reference) -> Reference | None:
        """Get the next chapter to the one containing this reference.

        This loops over the end of books using `refspy.book.Book.chapters`.
        This does not loop beyond the end of libraries.
        """
        return self.navigator.next_chapter(ref)

    def prev_chapter(self, ref: Reference) -> Reference | None:
        """Get the previous chapter to the one containing this reference.

        This loops over the end of books using `refspy.book.Book.chapters`, but
        not beyond the end of libraries.
        """
        return self.navigator.prev_chapter(ref)

    # -----------------------------------
    # Manipulation functions
    # -----------------------------------

    def get_book(self, ref: Reference) -> Book:
        """Get the book object for this reference's first range."""
        v1 = ref.ranges[0].start
        return self.books[v1.library, v1.book]

    def book_reference(self, ref: Reference) -> Reference:
        """Create a reference to the book containing this reference's first
        range."""
        v1 = ref.ranges[0].start
        return reference(
            range(
                verse(v1.library, v1.book, 1, 1),
                verse(v1.library, v1.book, 999, 999),
            )
        )

    def chapter_reference(self, ref: Reference) -> Reference:
        """Create a reference to the chapter containing this reference's first
        range."""
        v1 = ref.ranges[0].start
        return reference(
            range(
                verse(v1.library, v1.book, v1.chapter, 1),
                verse(v1.library, v1.book, v1.chapter, 999),
            )
        )

    # -----------------------------------
    # Formatting functions
    # -----------------------------------

    def name(self, ref: Reference) -> str:
        """Format a reference."""
        return self.formatter.format(ref, NAME_FORMAT)

    def book(self, ref: Reference) -> str:
        """Format a reference using only the book part of its name."""
        return self.formatter.format(ref, BOOK_FORMAT)

    def abbrev_name(self, ref: Reference) -> str:
        """Format an abbreviated reference."""
        return self.formatter.format(ref, ABBREV_NAME_FORMAT)

    def abbrev_book(self, ref: Reference) -> str:
        """Format an abbreviated reference using only the book part of its name."""
        return self.formatter.format(ref, ABBREV_BOOK_FORMAT)

    def numbers(self, ref: Reference) -> str:
        """Format a reference using only the number part of its name.

        The number part is the same for full names and abreviated names.
        """
        return self.formatter.format(ref, NUMBER_FORMAT)

    # -----------------------------------
    # Template formatting functions
    # -----------------------------------

    def template(self, reference: Reference, pattern: str) -> str:
        """
        Substitute formatting values in a string:

            * `{NAME}` -> "1 Corinthians 2:3–4"
            * `{BOOK}` -> "1 Corinthians"
            * `{NUMBERS}` -> "2:3–4"
            * `{ABBREV_NAME}` -> "1 Cor 2:3–4"
            * `{ABBREV_BOOK}` -> "1 Cor"
            * `{ABBREV_NUMBERS}` -> "2:3–4"
            * `{ESC_NAME}` -> "1%20Corinthians%202%3A3-4"
            * `{ESC_BOOK}` -> "1%20Corinthians"
            * `{ESC_NUMBERS}` -> "2%3A3-4"
            * `{ESC_ABBREV_NAME}` -> "1%20Cor%202%3A3-4"
            * `{ESC_ABBREV_BOOK}` -> "1%20Cor"
            * `{ESC_ABBREV_NUMBERS}` -> "2%3A3-4"
            * `{PARAM_NAME}`-> "1cor+2.3-4"
            * `{PARAM_BOOK}` -> "1cor"
            * `{PARAM_NUMBERS}` -> "2.3-4"

        We calculate only the values required by the template string.
        """
        out = pattern
        regexp = re.compile(r"\{[A-Z_]+\}")
        matches = regexp.findall(pattern)
        numbers = self.numbers(reference)
        for _ in matches:
            if _ == "{NAME}":
                out = out.replace("{NAME}", self.name(reference))
            elif _ == "{BOOK}":
                out = out.replace("{BOOK}", self.book(reference))
            elif _ == "{NUMBERS}":
                out = out.replace("{NUMBERS}", numbers)
            elif _ == "{ASCII_NUMBERS}":
                out = out.replace("{ASCII_NUMBERS}", numbers.replace("–", "-"))
            elif _ == "{ABBREV_NAME}":
                out = out.replace("{ABBREV_NAME}", self.abbrev_name(reference))
            elif _ == "{ABBREV_BOOK}":
                out = out.replace("{ABBREV_BOOK}", self.abbrev_name(reference))
            elif _ == "{ESC_NAME}":
                out = out.replace("{ESC_NAME}", url_escape(self.name(reference)))
            elif _ == "{ESC_BOOK}":
                out = out.replace("{ESC_BOOK}", url_escape(self.book(reference)))
            elif _ == "{ESC_NUMBERS}":
                out = out.replace("{ESC_NUMBERS}", url_escape(numbers))
            elif _ == "{ESC_ASCII_NUMBERS}":
                out = out.replace(
                    "{ESC_ASCII_NUMBERS}",
                    url_escape(numbers.replace("–", "-")),
                )
            elif _ == "{ESC_ABBREV_NAME}":
                out = out.replace(
                    "{ESC_ABBREV_NAME}", url_escape(self.abbrev_name(reference))
                )
            elif _ == "{ESC_ABBREV_BOOK}":
                out = out.replace(
                    "{ESC_ABBREV_BOOK}",
                    url_escape(self.abbrev_book(reference)),
                )
            elif _ == "{PARAM_NAME}":
                out = out.replace("{PARAM_NAME}", url_param(self.name(reference)))
            elif _ == "{PARAM_BOOK}":
                out = out.replace("{PARAM_BOOK}", url_param(numbers))
            elif _ == "{PARAM_NUMBERS}":
                out = out.replace("{PARAM_NUMBERS}", url_param(numbers))
        return out



from context import *

from pydantic import ValidationError
import pytest

from refspy.range import (
    Range,
    book_range,
    chapter_range,
    combine,
    merge,
    range,
    verse_range,
)
from refspy.verse import verse


def test_shorthand():
    v1 = verse(1, 2, 3, 4)
    v2 = verse(1, 2, 3, 5)
    _ = range(v1, v2)
    assert _.start == v1
    assert _.end == v2


def test_construction():
    v1 = verse(1, 2, 3, 4)
    v2 = verse(1, 2, 3, 5)
    _ = Range(start=v1, end=v2)
    assert _.start == v1
    assert _.end == v2


def test_wrong_verse_order():
    with pytest.raises(ValidationError):
        v1 = verse(1, 2, 3, 4)
        v2 = verse(1, 2, 3, 5)
        _ = Range(start=v2, end=v1)


def test_overlaps():
    r1 = verse_range(1, 2, 3, 4, 8)
    r2 = verse_range(1, 2, 3, 6, 8)
    r3 = verse_range(1, 2, 3, 9, 12)
    assert r1.overlaps(r2)
    assert not r1.overlaps(r3)
    assert not r2.overlaps(r3)


def test_contains():
    r1 = verse_range(1, 2, 3, 4, 8)
    r2 = verse_range(1, 2, 3, 6, 8)
    r3 = verse_range(1, 2, 3, 9, 12)
    assert r1.contains(r2)
    assert not r2.contains(r1)
    assert not r3.contains(r1)


def test_adjoins():
    vr1 = verse_range(1, 2, 3, 3, 4)
    vr2 = verse_range(1, 2, 3, 5, 6)
    cr1 = chapter_range(1, 2, 3, 4)
    cr2 = chapter_range(1, 2, 5, 6)
    assert vr1.adjoins(vr2)
    assert vr2.adjoins(vr1)
    assert cr1.adjoins(cr2)
    assert cr2.adjoins(cr1)
    assert not cr1.adjoins(vr1)
    assert not cr2.adjoins(vr2)


def test_adjoins_chapters():
    range_1 = chapter_range(1, 2, 3)
    range_2 = chapter_range(1, 2, 4, 5)
    assert range_1.adjoins(range_2)


def test_sorted():
    range_1 = verse_range(1, 2, 3, 4, 6)
    range_2 = verse_range(1, 2, 3, 6, 8)
    assert sorted([range_2, range_1]) == [range_1, range_2]


def test_merge_overlapping():
    range_1 = verse_range(1, 2, 3, 4, 6)
    range_2 = verse_range(1, 2, 3, 6, 8)
    range_3 = verse_range(1, 2, 3, 4, 8)
    assert merge([range_2, range_1]) == [range_3]


def test_merge_overlapping_inter_chapter():
    range_1 = verse_range(1, 2, 3, 4, 18)
    range_2 = range(verse(1, 2, 3, 11), verse(1, 2, 5, 2))
    result_1 = range(verse(1, 2, 3, 4), verse(1, 2, 5, 2))
    assert merge([range_1, range_2]) == [result_1]


def test_combine_adjacent_verses():
    range_1 = verse_range(1, 2, 3, 4, 6)
    range_2 = verse_range(1, 2, 3, 7, 8)
    range_3 = verse_range(1, 2, 3, 4, 8)
    assert combine([range_2, range_1]) == [range_3]


def test_combine_adjacent_verses_complex():
    range_1 = verse_range(1, 2, 3, 4, 6)
    range_2 = verse_range(1, 2, 3, 9, 12)
    range_3 = verse_range(1, 2, 3, 11, 15)
    range_4 = verse_range(1, 2, 3, 2, 5)
    result_1 = verse_range(1, 2, 3, 2, 6)
    result_2 = verse_range(1, 2, 3, 9, 15)
    assert combine([range_1, range_2, range_3, range_4]) == [result_1, result_2]


def test_combine_adjacent_verses_inter_chapter():
    range_1 = range(verse(1, 2, 3, 4), verse(1, 2, 4, 18))
    range_2 = range(verse(1, 2, 4, 11), verse(1, 2, 5, 2))
    result_1 = range(verse(1, 2, 3, 4), verse(1, 2, 5, 2))
    assert combine([range_1, range_2]) == [result_1]


def test_combine_adjacent_chapters():
    range_1 = chapter_range(1, 2, 3)
    range_2 = chapter_range(1, 2, 4, 5)
    combined = combine([range_2, range_1])
    assert combined == [range(verse(1, 2, 3, 1), verse(1, 2, 5, 999))]


def test_combine_adjacent_books():
    range_1 = book_range(1, 2, 3)
    range_2 = book_range(1, 4, 5)
    combined = combine([range_2, range_1])
    assert combined == [range(verse(1, 2, 1, 1), verse(1, 5, 999, 999))]


def test_is_book_range():
    assert book_range(1, 2, 3).is_book_range()


def test_is_inter_book_range():
    assert range(verse(1, 1, 1, 1), verse(1, 2, 1, 1)).is_inter_book_range()


def test_is_book():
    assert book_range(1, 2).is_book()


def test_is_chapter_range():
    assert chapter_range(1, 2, 3, 4).is_chapter_range()


def test_is_inter_chapter_range():
    assert range(verse(1, 1, 1, 2), verse(1, 1, 3, 4)).is_inter_chapter_range()


def test_is_chapter():
    assert chapter_range(1, 2, 3).is_chapter()


def test_is_verse_range():
    assert verse_range(1, 2, 3, 4, 5).is_verse_range()


def test_is_verse():
    assert verse_range(1, 2, 3, 4).is_verse()
